# step 步驟
# 建立 interface 介面
from abc import ABC, abstractmethod  # Abstract Base Class 抽象類別


class Step(ABC):  # Base class
    def __init__(self):
        pass

    @abstractmethod
    def process(self, transporter, inputs, utils):  # 執行
        pass


class StepException(Exception):  # 例外捕捉
    pass



import sys
import getopt

from yt_concate.pipeline.pipeline import Pipeline
from yt_concate.pipeline.steps.preflight import Preflight
from yt_concate.pipeline.steps.get_v_links import GetVideoList
from yt_concate.pipeline.steps.initialize_yto import InitializeYTO
from yt_concate.pipeline.steps.download_captions import DownloadCaptions
from yt_concate.pipeline.steps.read_captions import ReadCaptions
from yt_concate.pipeline.steps.search import Search
from yt_concate.pipeline.steps.download_video import DownLoadVideos
from yt_concate.pipeline.steps.edit_video import EditVideo
from yt_concate.pipeline.steps.postflight import Postflight
from yt_concate.utils import Utils

CHANNEL_ID = 'UCKSVUHI9rbbkXhvAXK-2uxA'
SEARCH_WORD = 'incredible'
# 輸入字典
inputs = {
    'channel_id': CHANNEL_ID,
    'search_word': SEARCH_WORD,
    'limit': 10,
}


def print_usage():
    print('說明書在此:')
    print('python yt-concat.py OPERATIONS')
    print('OPERATION:')
    print('{:>6}  {:<16}{}'.format('-c', '--channel_id', 'channel_id for video channel'))
    print('{:>6}  {:<16}{}'.format('-s', '--search_word',
                                   '--search_word for the word you want to search in channel video'))
    print('{:>6}  {:<16}{}'.format('-l', '--limit', 'limit for download video clips'))


def main():
    # 步驟清單，裝著所有步驟，讓底下的for loop一個一個去call
    steps = [
        Preflight(),
        GetVideoList(),
        InitializeYTO(),
        DownloadCaptions(),
        ReadCaptions(),
        Search(),
        DownLoadVideos(),
        EditVideo(),
        Postflight(),
    ]

    # Command Line Arguments 程式參數化
    short_opts = 'hc:s:l:'
    long_opts = 'help channel_id= search_word= limit='.split()
    try:
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError:
        print_usage()
        sys.exit(2)

    for opt, arg in opts:
        if opt == '-h':
            print_usage()
            sys.exit(0)
        elif opt in ("-c", "--channel_id"):
            inputs['channel_id'] = arg
        elif opt in ("-s", "--search_word"):
            inputs['search_word'] = arg
        elif opt in ("-l", "--limit"):
            inputs['limit'] = int(arg)

    utils = Utils()
    p = Pipeline(steps)
    p.run(inputs, utils)


if __name__ == '__main__':
    main()



# pipeline 生產線
# 裝載著所有步驟
from yt_concate.pipeline.steps.step import StepException


class Pipeline:
    def __init__(self, steps):
        self.steps = steps

    def run(self, inputs, utils):
        transporter = None  # 運輸車，把生產線的東西一個一個傳給亞一個生產線，初始為None代表沒有東西。
        for work in self.steps:
            try:
                transporter = work.process(transporter, inputs, utils)
            except StepException as err:
                print('休士頓，我們發現了一個錯誤！', err)
                break



from __future__ import unicode_literals
import youtube_dl
import os
from yt_concate.pipeline.steps.step import Step


class DownLoadVideos(Step):

    def process(self, transporter, inputs, utils):
        counter = 0
        for catch_yto in transporter:
            if counter < inputs['limit']:
                print('第', counter + 1, '支影片開始下載')
                if utils.videos_file_exists(catch_yto.yto):
                    print('此影片已下載', catch_yto.yto.v_id)
                    counter += 1
                    continue
                save_path = '/'.join(os.getcwd().split('/')[:3]) + '/downloads/videos'
                file_name = f'/{catch_yto.yto.v_id}.mp4'

                # yortube_id operations，yortube_id的各項屬性，可以根據你的需求加減
                ydl_opts = {
                    'format': 'best',  # 畫質：最差 (為預防我的C槽炸裂)
                    'outtmpl': save_path + file_name,  # 輸出模板：裝著路徑以及檔名
                }

                # 記得YoutubeDL()裡面裝著ydl_opt 以及download([url]) 是小誇號裝著中誇號
                youtube_dl.YoutubeDL(ydl_opts).download([catch_yto.yto.url])
                counter += 1
        print('影片下載完成')
        return transporter



from yt_concate.pipeline.steps.step import Step
from yt_concate.model.catch_yto import CatchYTO

class Search(Step):
    def process(self, transporter, inputs, utils):
        found = []
        counter = 0
        for yto in transporter:
            n = 0
            while True:
                try:
                    if inputs['search_word'] in yto.captions[n]['text']:
                        time_start = yto.captions[n]['start']
                        time_duration = yto.captions[n]['duration']
                        c = CatchYTO(yto, time_start, time_duration)
                        found.append(c)
                        counter += 1
                        break
                except IndexError:
                    break
                except TypeError:
                    print('格式錯誤', yto.url)
                    break
                n += 1

        print('搜索任務完成')
        print('共找到', counter, '個影片有', inputs['search_word'])
        return found



import os
from dotenv import load_dotenv

load_dotenv()  # take environment variables from .env.

API_KEY = os.getenv('API_KEY')

# 把CAP跟VID加進DOW裡面
DOWNLOADS_DIR = 'downloads'  # 下載資料夾
VIDEOS_DIR = os.path.join(DOWNLOADS_DIR, 'videos')  # 影片資料夾
CAPTIONS_DIR = os.path.join(DOWNLOADS_DIR, 'captions')  # 字幕資料夾
OUTPUT_DIR = 'output'



# 下載字幕
from youtube_transcript_api import YouTubeTranscriptApi
from yt_concate.pipeline.steps.step import Step
import pickle


class DownloadCaptions(Step):
    def process(self, transporter, inputs, utils):
        print('下載字幕中')
        for yto in transporter:
            if utils.caption_file_exists(yto):
                print('已經下載過了喔:', yto.v_id)
                continue
            try:
                # 使用 srt 變量和 .get_transcript() 函數獲得的字典列表
                srt = YouTubeTranscriptApi.get_transcript(yto.v_id)
            except:
                print('非英文或無字幕，網址: ', yto.url)
                continue
            print('寫入中', yto.url)
            with open(yto.get_caption_filepath(), 'wb', ) as f:
                pickle.dump(srt, f)
        return transporter



from yt_concate.pipeline.steps.step import Step
import pickle


class ReadCaptions(Step):
    def process(self, transporter, inputs, utils):
        print('處理字幕中......')
        for yto in transporter:
            try:
                with open(yto.caption_filepath, 'rb') as f:
                    yto.captions = pickle.load(f)
            except FileNotFoundError:
                print('未找到此檔案', yto.v_id)
                continue
        return transporter



# utils 實用程式，屬於工具箱，裝著一些通用型小型def
import os
from yt_concate.settings import DOWNLOADS_DIR, CAPTIONS_DIR, VIDEOS_DIR, OUTPUT_DIR


class Utils:
    def __init__(self):
        pass

    def create_dir(self):
        os.makedirs(DOWNLOADS_DIR, exist_ok=True)
        os.makedirs(CAPTIONS_DIR, exist_ok=True)
        os.makedirs(VIDEOS_DIR, exist_ok=True)
        os.makedirs(OUTPUT_DIR, exist_ok=True)

    def get_videos_list_filepath(self, channel_id):
        return os.path.join(DOWNLOADS_DIR, channel_id + '.txt')

    def videos_list_file_exists(self, channel_id):
        path = self.get_videos_list_filepath(channel_id)
        return os.path.exists(path) and os.path.getsize(path) > 0

    def caption_file_exists(self, yto):
        filepath = yto.get_caption_filepath()
        return os.path.exists(filepath) and os.path.getsize(filepath) > 0

    def videos_file_exists(self, yto):
        filepath = yto.get_video_filepath()
        return os.path.exists(filepath) and os.path.getsize(filepath) > 0

    def edit_videos_file_exists(self, inputs):
        filepath = self.get_output_filename(inputs['channel_id'], inputs['search_word'])
        return os.path.exists(filepath) and os.path.getsize(filepath) > 0

    def get_output_filename(self, channel_id, search_word):
        filename = channel_id + '_' + search_word + '.mp4'
        return os.path.join(OUTPUT_DIR, filename)



from .step import Step
from yt_concate.model.yto import YTO


class InitializeYTO(Step):
    def process(self, transporter, inputs, utils):
        print('初始化處理中......')
        return [YTO(url) for url in transporter]



from yt_concate.pipeline.steps.step import Step
from moviepy.editor import VideoFileClip, concatenate_videoclips


class EditVideo(Step):
    def process(self, transporter, inputs, utils):
        if utils.edit_videos_file_exists(inputs):
            print('影片已編輯完成')
            return
        print('影片開始製作')
        n = 1
        clips = []
        for catch_yto in transporter:
            start = catch_yto.time_start
            end = round(catch_yto.time_end, 3)
            print('第', n, '隻影片加入剪輯區')
            n += 1
            video = VideoFileClip(catch_yto.yto.video_filepath).subclip(start, end)
            clips.append(video)
            if len(clips) >= inputs['limit']:
                break
        print('開始剪輯')
        final_clip = concatenate_videoclips(clips)
        final_clip.write_videofile(utils.get_output_filename(inputs['channel_id'], inputs['search_word']))
        print('剪輯結束')
